from qiskit import Aer, QuantumCircuit, execute
from qiskit.compiler import transpile, assemble
from quantastica.qiskit_toaster import ToasterBackend
import time
import logging
import os

backend = ToasterBackend.get_backend("qasm_simulator")
# backend = Aer.get_backend("qasm_simulator")
default_options = {
    # "method": "statevector",   # Force dense statevector method for benchmarks
    # "truncate_enable": False,  # Disable unused qubit truncation for benchmarks
    # "max_parallel_threads": 1,  # Disable OpenMP parallelization for benchmarks
    "toaster_optimization": 3  # optimization for qubit-toaster
}

# def _execute(circuit, backend_options=None):
#     experiment = transpile(circuit, backend)
#     qobj = assemble(experiment, shots=1)
#     qobj_aer = backend._format_qobj(qobj, backend_options, None)
#     return backend._controller(qobj_aer)


def _execute(circuit, backend):
    job = execute(circuit, backend=backend)
    # job.result is needed in order to wait for results
    job.result()


def native_execute(benchmark, circuit, backend_options=None):
    # experiment = transpile(circuit, backend)
    # qobj = assemble(experiment, shots=1)
    # qobj_aer = backend._format_qobj(qobj, backend_options, None)
    # benchmark(backend._controller, qobj_aer)
    benchmark(_execute, circuit, backend)


def run_bench(benchmark, nqubits, gate, locs=(1,)):
    qc = QuantumCircuit(nqubits)
    getattr(qc, gate)(*locs)
    native_execute(benchmark, qc, default_options)


def first_rotation(circuit, nqubits):
    circuit.rx(1.0, range(nqubits))
    circuit.rz(1.0, range(nqubits))
    return circuit


def mid_rotation(circuit, nqubits):
    circuit.rz(1.0, range(nqubits))
    circuit.rx(1.0, range(nqubits))
    circuit.rz(1.0, range(nqubits))
    return circuit


def last_rotation(circuit, nqubits):
    circuit.rz(1.0, range(nqubits))
    circuit.rx(1.0, range(nqubits))
    return circuit


def entangler(circuit, pairs):
    for a, b in pairs:
        circuit.cx(a, b)
    return circuit


def generate_qcbm_circuit(nqubits, depth, pairs):
    circuit = QuantumCircuit(nqubits)
    first_rotation(circuit, nqubits)
    entangler(circuit, pairs)
    for k in range(depth - 1):
        mid_rotation(circuit, nqubits)
        entangler(circuit, pairs)
    last_rotation(circuit, nqubits)
    # circuit.measure_all()
    return circuit


def test_qcbm(name, backend, nqubits, optimization_level=None):
    pairs = [(i, (i + 1) % nqubits) for i in range(nqubits)]
    circuit = generate_qcbm_circuit(nqubits, 9, pairs)
    with open("circuit.qasm", "w") as f:
        f.write(circuit.qasm())
    t1 = time.time()

    # test = transpile(circuit, optimization_level=0)
    # qobj = assemble(circuit)
    # job = backend.run(qobj, backend_options=default_options)

    job = execute(
        circuit,
        backend=backend,
        backend_options=default_options,
        optimization_level=optimization_level,
    )

    # job.result is needed in order to wait for results
    result = job.result()
    t2 = time.time()
    # print(result)
    # counts = result.get_counts(circuit)
    # print("Counts size: %d" % len(counts))
    print("  %s time: %f" % (name, t2 - t1))


if __name__ == "__main__":
    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(pathname)s - %(message)s",
        level=os.environ.get("LOGLEVEL", "CRITICAL"),
    )
    N = int(os.environ.get("N", 20))
    LOOPS = int(os.environ.get("LOOPS", 10))
    for i in range(0, LOOPS):
        print("Iteration #%d (of %d)" % (i, LOOPS))
        test_qcbm("AER", Aer.get_backend("qasm_simulator"), N, 0)
        test_qcbm(
            "TOASTER", ToasterBackend.get_backend("qasm_simulator"), N, 0
        )



# This code is part of quantastica.qiskit_toaster
#
# (C) Copyright Quantastica 2019.
# https://quantastica.com/
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
import logging
import subprocess

logger = logging.getLogger(__name__)


class ToasterCliInterface:
    def __init__(self, toaster_path):
        self.toaster_path = toaster_path

    def execute(
        self,
        jsonstr,
        job_id=None,
        seed=None,
        shots=None,
        returns=None,
        optimization=None,
    ):

        args = [self.toaster_path, "-", "-s", str(shots)]
        if returns:
            returns = returns.split(",")
        else:
            returns = ["counts"]
        args.append("-r")
        args += returns

        if seed:
            args.append("--seed")
            args.append(str(seed))
        if optimization:
            args.append("-o")
            args.append(str(optimization))

        proc = subprocess.Popen(
            args,
            close_fds=False,
            restore_signals=False,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        logger.info("Running q-toaster with following params:")
        logger.info(args)
        proc.stdin.write(jsonstr)
        proc.stdin.close()
        proc.wait()
        qtoasterjson = proc.stdout.read()
        proc.stdout.close()
        stderr = proc.stderr.read()
        proc.stderr.close()
        returncode = proc.returncode
        if returncode > 0:
            logger.debug(
                "Toaster finished with non-zero exit code (%d) :" % returncode,
                stderr,
            )
            raise RuntimeError(
                "Error received from CLI, exit code: %d" % returncode
            )

        return qtoasterjson



# This code is part of quantastica.qiskit_toaster
#
# (C) Copyright Quantastica 2019.
# https://quantastica.com/
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from concurrent import futures
import logging
import json
import time
import copy
import os
import sys

from quantastica.qconvert import qobj_to_toaster
from quantastica.qiskit_toaster import (
    ToasterHttpInterface,
    ToasterCliInterface,
)

from qiskit.providers import BaseJob, JobStatus, JobError
from qiskit.result import Result

logger = logging.getLogger(__name__)


# one of toaster_url or toaster_path MUST be defined
# if both are defined toaster_path takes precedence
def _run_with_qtoaster_static(
    qobj_dict,
    get_states,
    job_id,
    optimization_level=None,
    toaster_url=None,
    toaster_path=None,
):
    SEED_SIMULATOR_KEY = "seed_simulator"
    if get_states:
        shots = 1
    else:
        shots = qobj_dict["config"]["shots"]

    returns = "counts"
    if get_states:
        returns += ",state"

    seed = 0
    if SEED_SIMULATOR_KEY in qobj_dict["config"]:
        seed = qobj_dict["config"][SEED_SIMULATOR_KEY]

    converted = qobj_to_toaster(qobj_dict, {"all_experiments": False})

    dump_dir = os.getenv("TOASTER_DUMP_DIR", None)
    if dump_dir is not None:
        path_req = "%s/%s.request.json" % (dump_dir, job_id)
        with open(path_req, "w") as f:
            f.write(converted)
    if toaster_path:
        toaster = ToasterCliInterface.ToasterCliInterface(toaster_path)
    else:
        toaster = ToasterHttpInterface.ToasterHttpInterface(toaster_url)

    toasterjson = toaster.execute(
        converted.encode("utf-8"),
        job_id=job_id,
        returns=returns,
        seed=seed,
        optimization=optimization_level,
        shots=shots,
    )
    if dump_dir is not None:
        path_res = "%s/%s.response.json" % (dump_dir, job_id)
        with open(path_res, "w") as f:
            f.write(str(toasterjson))

    resultraw = None
    if toasterjson:
        resultraw = json.loads(toasterjson)

    success = resultraw is not None
    # print(success)
    data = dict()
    exp_dict = qobj_dict["experiments"][0]
    exp_header = exp_dict["header"]
    expname = exp_header["name"]
    time_taken = 0
    rawversion = "0.0.0"
    if success:
        logger.debug("Raw results from toaster:\r\n%s", resultraw)
        if isinstance(resultraw, dict):
            rawversion = resultraw.get("qtoaster_version")

        if ToasterJob._check_qtoaster_version(rawversion) is False:
            raise ValueError(
                "Unsupported qtoaster_version, got '%s' - minimum expected is '%s'.\n\rPlease update your q-toaster to latest version"
                % (rawversion, ToasterJob._MINQTOASTERVERSION)
            )
        counts = ToasterJob._convert_counts(resultraw["counts"])
        statevector = resultraw.get("statevector")
        data["counts"] = counts
        if statevector is not None and len(statevector) > 0:
            data["statevector"] = statevector
        time_taken = resultraw["time_taken"]

    result = {
        "success": success,
        "meas_level": 2,
        "shots": shots,
        "data": data,
        "header": exp_header,
        "status": "DONE",
        "time_taken": time_taken,
        "name": expname,
        "seed_simulator": seed,
        "toaster_version": rawversion,
    }
    return result


class ToasterJob(BaseJob):
    DEFAULT_TOASTER_HOST = "127.0.0.1"
    DEFAULT_TOASTER_PORT = 8001
    _MINQTOASTERVERSION = "0.9.9"

    if sys.platform in ["darwin", "win32"]:
        _executor = futures.ThreadPoolExecutor(max_workers=2)
    else:
        _executor = futures.ProcessPoolExecutor(max_workers=2)
    _run_time = 0

    def __init__(
        self,
        backend,
        job_id,
        qobj,
        toaster_host,
        toaster_port,
        getstates=False,
        backend_options=None,
        use_cli=False,
    ):
        super().__init__(backend, job_id)
        self._toaster_url = "http://%s:%d" % (toaster_host, int(toaster_port))
        self._result = None
        self._qobj_dict = qobj.to_dict()
        self._futures = []
        self._getstates = getstates
        self._backend_options = backend_options
        self._use_cli = use_cli

    def submit(self):
        if len(self._futures) > 0:
            raise JobError("We have already submitted the job!")
        self._t_submit = time.time()

        logger.debug("submitting...")
        all_exps = self._qobj_dict
        exp_index = 0
        optimization_level = None
        backend_options = self._backend_options
        if backend_options:
            optimization_level = backend_options.get(
                "toaster_optimization", None
            )

        toaster_path = None
        if int(self._use_cli) != 0:
            toaster_path = "qubit-toaster"

        for exp in all_exps["experiments"]:
            exp_index += 1
            exp_job_id = "Exp_%d_%s" % (exp_index, self._job_id)
            single_exp = copy.deepcopy(all_exps)
            single_exp["experiments"] = [exp]
            self._futures.append(
                self._executor.submit(
                    _run_with_qtoaster_static,
                    single_exp,
                    self._getstates,
                    exp_job_id,
                    optimization_level=optimization_level,
                    toaster_url=self._toaster_url,
                    toaster_path=toaster_path,
                )
            )

    def wait(self, timeout=None):
        if self.status() in [JobStatus.RUNNING, JobStatus.QUEUED]:
            futures.wait(self._futures, timeout)
        if self._result is None and self.status() is JobStatus.DONE:
            results = []
            for f in self._futures:
                results.append(f.result())
            qobj_dict = self._qobj_dict
            qobjid = qobj_dict["qobj_id"]
            qobj_header = qobj_dict["header"]
            rawversion = "1.0.0"
            if len(results):
                if "toaster_version" in results[0]:
                    rawversion = results[0]["toaster_version"]

            self._result = {
                "success": True,
                "backend_name": "Toaster",
                "qobj_id": qobjid,
                "backend_version": rawversion,
                "header": qobj_header,
                "job_id": self._job_id,
                "results": results,
                "status": "COMPLETED",
            }
            ToasterJob._run_time += time.time() - self._t_submit

        if len(self._futures) > 0:
            for f in self._futures:
                if f.exception():
                    raise f.exception()

    def result(self, timeout=None):
        self.wait(timeout)
        return Result.from_dict(self._result)

    def cancel(self):
        return
        # return self._future.cancel()

    def status(self):

        if len(self._futures) == 0:
            _status = JobStatus.INITIALIZING
        else:
            running = 0
            done = 0
            canceled = 0
            error = 0
            queued = 0
            for f in self._futures:
                if f.running():
                    running += 1
                elif f.cancelled():
                    canceled += 1
                elif f.done():
                    if f.exception() is None:
                        done += 1
                    else:
                        error += 1
                else:
                    queued += 1

            if error:
                _status = JobStatus.ERROR
            elif running:
                _status = JobStatus.RUNNING
            elif canceled:
                _status = JobStatus.CANCELLED
            elif done:
                _status = JobStatus.DONE
            else:  # future is in pending state
                _status = JobStatus.QUEUED
        return _status

    def backend(self):
        """Return the instance of the backend used for this job."""
        return self._backend

    @staticmethod
    def _qtoaster_version_to_int(versionstring):
        parts = versionstring.split(".")
        verint = 0
        if len(parts) >= 3:
            verint = (
                int(parts[0]) * 1000000 + int(parts[1]) * 1000 + int(parts[2])
            )
        return verint

    @classmethod
    def _check_qtoaster_version(cls, versionstring):
        verint = cls._qtoaster_version_to_int(versionstring)
        minverint = cls._qtoaster_version_to_int(cls._MINQTOASTERVERSION)
        if verint < minverint:
            return False
        return True

    @staticmethod
    def _convert_counts(counts):
        ret = dict()
        for key in counts:
            nicekey = key.replace(" ", "")
            nicekey = hex(int(nicekey, 2))
            ret[nicekey] = counts[key]
        return ret



import os
from quantastica.qiskit_toaster import ToasterBackend
import logging
import unittest
import sys
import time


class TestToasterBase(unittest.TestCase):
    @staticmethod
    def toaster_backend(backend_name=None):
        return ToasterBackend.get_backend(
            backend_name,
            toaster_host=os.getenv("TOASTER_HOST", None),
            toaster_port=os.getenv("TOASTER_PORT", None),
            use_cli=os.getenv("USE_CLI", False),
        )

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(
            format="%(levelname)s %(asctime)s %(pathname)s - %(message)s",
            level=os.environ.get("LOGLEVEL", "CRITICAL"),
        )

    def setUp(self):
        self.startTime = time.time()

    def tearDown(self):
        t = time.time() - self.startTime
        sys.stderr.write(" took %.3fs ... " % (t))



# This code is part of quantastica.qiskit_toaster
#
# (C) Copyright Quantastica 2019.
# https://quantastica.com/
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
import logging
import urllib
from urllib import request
import time
import socket


logger = logging.getLogger(__name__)


class ToasterHttpInterface:
    def __init__(self, toaster_url):
        self.toaster_url = toaster_url

    def execute(
        self,
        jsonstr,
        job_id=None,
        seed=None,
        shots=None,
        returns=None,
        optimization=None,
    ):

        params = dict()
        params["x-qtc-return"] = returns or "counts"
        params["x-qtc-shots"] = "%d" % (shots or 1)
        params["x-qtc-jobid"] = job_id or ""
        if seed:
            params["x-qtc-seed"] = "%d" % seed
        if optimization:
            params["x-qtc-optimization"] = "%d" % optimization

        logger.info("Sending circuit to toaster, url: %s", self.toaster_url)
        logger.info("Simulation params: %s", params)
        timeout = None
        params["content-type"] = "application/json"
        req = request.Request(
            self.toaster_url + "/submit", data=jsonstr, headers=params
        )
        txt = None
        max_retries = 5
        retry_count = 0

        while True:
            try:
                response = request.urlopen(req, timeout=timeout)
            except socket.timeout as e:
                logger.debug("Exception raised: %s", e)
                txt = self._fetch_last_response(timeout, job_id)
                if txt is None:
                    continue
                else:
                    break
            except urllib.error.HTTPError as e:
                logger.debug("Exception raised: %s", e)
                # already submitted, lets fetch results
                if e.code == 409:
                    txt = self._fetch_last_response(timeout, job_id)
                    if txt is None:
                        continue
                    else:
                        break
                else:
                    raise RuntimeError(
                        "Error received from API(2): %s" % str(e)
                    )
            except Exception:
                if retry_count < max_retries:
                    retry_count += 1
                    logger.debug(
                        "Connection failed, retrying (#%d)...", retry_count
                    )
                    time.sleep(0.2)
                else:
                    msg = (
                        "Failed to connect to qubit-toaster, probably not running (url: %s)"
                        % self.toaster_url
                    )
                    logger.critical(msg)
                    raise RuntimeError(msg)
            else:
                txt = response.read().decode("utf8")
                break

        return txt

    def _fetch_last_response(self, timeout, job_id):
        req = request.Request("%s/pollresult/%s" % (self.toaster_url, job_id))
        txt = None
        while True:
            try:
                response = request.urlopen(req, timeout=timeout)
            except urllib.error.HTTPError as e:
                raise RuntimeError("Error received from API(1): %s" % str(e))
            except (socket.timeout, urllib.error.URLError) as e:
                logger.debug("Exception raised: %s", e)
                time.sleep(0.2)
                continue
            else:
                txt = response.read().decode("utf8")
                break

        return txt



# This code is part of quantastica.qiskit_toaster
#
# (C) Copyright Quantastica 2019.
# https://quantastica.com/
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import uuid
from quantastica.qiskit_toaster import ToasterJob
from qiskit.providers import BaseBackend
from qiskit.providers.models import BackendConfiguration


class ToasterBackend(BaseBackend):

    DEFAULT_TOASTER_HOST = ToasterJob.ToasterJob.DEFAULT_TOASTER_HOST
    DEFAULT_TOASTER_PORT = ToasterJob.ToasterJob.DEFAULT_TOASTER_PORT

    MAX_QUBITS_MEMORY = 32
    DEFAULT_CONFIGURATION = {
        "backend_name": "toaster_simulator",
        "backend_version": "1.0.0",
        "n_qubits": MAX_QUBITS_MEMORY,
        "url": "https://quantastica.com/",
        "simulator": True,
        "local": True,
        "conditional": False,
        "open_pulse": False,
        "memory": True,
        "max_shots": 65536,
        "description": "An q-toaster based qasm simulator",
        "coupling_map": None,
        "basis_gates": [
            "u1",
            "u2",
            "u3",
            "cx",
            "id",
            "x",
            "y",
            "z",
            "h",
            "s",
            "t",
        ],
        "gates": [],
    }

    def __init__(
        self,
        configuration=None,
        provider=None,
        backend_name=None,
        toaster_host=None,
        toaster_port=None,
        use_cli=False,
    ):
        configuration = configuration or BackendConfiguration.from_dict(
            self.DEFAULT_CONFIGURATION
        )

        # disable unroller
        # configuration.basis_gates=None

        super().__init__(configuration=configuration, provider=provider)

        getstates = False
        if (
            backend_name is not None
            and backend_name == "statevector_simulator"
        ):
            getstates = True

        self._getstates = getstates
        self._toaster_port = (
            toaster_port or ToasterBackend.DEFAULT_TOASTER_PORT
        )
        self._toaster_host = (
            toaster_host or ToasterBackend.DEFAULT_TOASTER_HOST
        )
        self._use_cli = use_cli

    # @profile
    def run(self, qobj, backend_options=None):
        job_id = str(uuid.uuid4())
        job = ToasterJob.ToasterJob(
            self,
            job_id,
            qobj,
            getstates=self._getstates,
            toaster_host=self._toaster_host,
            toaster_port=self._toaster_port,
            backend_options=backend_options,
            use_cli=self._use_cli,
        )
        job.submit()
        return job

    @staticmethod
    def name():
        return "qubit_toaster"


def get_backend(
    backend_name=None, toaster_host=None, toaster_port=None, use_cli=False
):
    return ToasterBackend(
        backend_name=backend_name,
        toaster_host=toaster_host,
        toaster_port=toaster_port,
        use_cli=use_cli,
    )
